//+------------------------------------------------------------------+
//|                                                      OLAPQTS.mq5 |
//|                                 Copyright © 2016-2020, Marketeer |
//|                          https://www.mql5.com/en/users/marketeer |
//|               Online Analytical Processing of trading hypercubes |
//|                            https://www.mql5.com/ru/articles/6602 |
//|                            https://www.mql5.com/ru/articles/6603 |
//|                            https://www.mql5.com/ru/articles/7535 |
//+------------------------------------------------------------------+

#property copyright "Copyright © 2016-2020, Marketeer"
#property link      "https://www.mql5.com/ru/articles/7535"
#property version   "1.0"
#property description "Quote History Analysis: Non-trading Expert Adviser demonstrating OLAP classes"


// forward declaration of our custom class
class CustomQuotesRecord;

// this line plugs our class into default adapter in OLAPQuotes.mqh
#define RECORD_CLASS CustomQuotesRecord


#include <OLAP/OLAPQuotes.mqh>
#include <Marketeer/GroupSettings.mqh>


input GroupSettings OLAP_Settings; // O L A P     S E T T I N G S

sinput string X = "————— X axis —————"; // · X ·
input SELECTORS SelectorX = DEFAULT_SELECTOR_TYPE; // · SelectorX
input ENUM_FIELDS FieldX = DEFAULT_SELECTOR_FIELD /* field does matter only for some selectors */; // · FieldX

sinput string Y = "————— Y axis —————"; // · Y ·
input SELECTORS SelectorY = SELECTOR_NONE; // · SelectorY
input ENUM_FIELDS FieldY = FIELD_NONE; // · FieldY

sinput string Z = "————— Z axis —————"; // · Z ·
input SELECTORS SelectorZ = SELECTOR_NONE; // · SelectorZ
input ENUM_FIELDS FieldZ = FIELD_NONE; // · FieldZ

sinput string F = "————— Filter —————"; // · F ·
input SELECTORS _Filter1 = SELECTOR_NONE; // · Filter1
input ENUM_FIELDS _Filter1Field = FIELD_NONE; // · Filter1Field
input string _Filter1value1 = ""; // · Filter1value1
input string _Filter1value2 = ""; // · Filter1value2

sinput string A = "————— Aggregator —————"; // · A ·
input AGGREGATORS _AggregatorType = DEFAULT_AGGREGATOR_TYPE; // · AggregatorType
input ENUM_FIELDS _AggregatorField = DEFAULT_AGGREGATOR_FIELD; // · AggregatorField
input SORT_BY _SortBy = SORT_BY_NONE; // · SortBy

input uint QuantGranularity = 0; // · QuantGranularity (0 - natural buckets)
// if specified, can be used for e.g. price ranges in points, profits in increments, etc.


SELECTORS _selectorArray[4];
ENUM_FIELDS _selectorField[4];


class CustomQuotesRecord: public QuotesRecord
{
  public:
    static string customFieldsTitles[4];

  private:
    static double previousBarRange;
    static double previousSpread;
    
    // Start-up time announcement of the custom fields specialization
    class CustomFieldsDescription
    {
      private:
        static CustomFieldsDescription instance;

        CustomFieldsDescription()
        {
          Print("Custom fields processing attached:");
          for(int i = FIELD_CUSTOM1; i <= FIELD_CUSTOM4; i++)
          {
            Print(EnumToString((QUOTE_RECORD_FIELDS)i), " == ", CustomQuotesRecord::customFieldsTitles[i - FIELD_CUSTOM1]);
          }
        }
    };

    int MathSignNonZero(const double x)
    {
      return x >= 0 ? +1 : -1;
    }

  public:
    CustomQuotesRecord(): QuotesRecord() {}
    CustomQuotesRecord(const MqlRates &rate): QuotesRecord(rate)
    {
    }
    
    virtual void fillCustomFields() override
    {
      const double newBarRange = get(FIELD_PRICE_RANGE_OC);
      const double spread = get(FIELD_SPREAD);

      if(MathAbs(previousBarRange) > previousSpread)
      {
        double mult = newBarRange * previousBarRange;
        double value = MathSign(mult) * MathAbs(newBarRange);

        // this is an attempt to approximate average losses due to spreads
        value += MathSignNonZero(value) * -1 * MathMax(spread, previousSpread);
        
        set(FIELD_CUSTOM3, value);
        set(FIELD_CUSTOM4, MathSign(mult));
      }
      else
      {
        set(FIELD_CUSTOM3, 0);
        set(FIELD_CUSTOM4, 0);
      }

      set(FIELD_CUSTOM1, MathSign(newBarRange) * (MathAbs(newBarRange) - spread));
      set(FIELD_CUSTOM2, MathSign(newBarRange) * MathSign(MathAbs(newBarRange) - spread));
      
      previousBarRange = newBarRange;
      previousSpread = spread;
    }
    
    virtual string legend(const int index) const override
    {
      if(index >= FIELD_CUSTOM1 && index <= FIELD_CUSTOM4)
      {
        return CustomQuotesRecord::customFieldsTitles[index - FIELD_CUSTOM1];
      }
      
      return QuotesRecord::legend(index);
    }
};

static double CustomQuotesRecord::previousBarRange = 0;
static double CustomQuotesRecord::previousSpread = 0;
static CustomQuotesRecord::CustomFieldsDescription CustomQuotesRecord::CustomFieldsDescription::instance;
static string CustomQuotesRecord::customFieldsTitles[4] =
{
  "Real range (w/o spread)",
  "Real direction counter (w/o spread)",
  "Next bar covariance",
  "Next bar direction counter"
};


LogDisplay _display(11, _Digits, true);

class MyProgress: public Progress
{
  public:
    void progress(const int cursor, const int total) const override
    {
      Comment((string)cursor + " / " + (string)total);
    }
};

MyProgress progress;


int OnInit()
{
  _selectorArray[0] = SelectorX;
  _selectorArray[1] = SelectorY;
  _selectorArray[2] = SelectorZ;
  _selectorArray[3] = _Filter1;
  _selectorField[0] = FieldX;
  _selectorField[1] = FieldY;
  _selectorField[2] = FieldZ;
  _selectorField[3] = _Filter1Field;
  
  _defaultEngine.setAdapter(&_defaultQuotesAdapter);
  _defaultEngine.setQuant(QuantGranularity);
  _defaultEngine.setProgress(&progress);

  EventSetTimer(1);
  return INIT_SUCCEEDED;
}

void OnTimer()
{
  double Filter1value1 = 0, Filter1value2 = 0;
  if(CustomQuotesRecord::datatype(_Filter1Field) == 't')
  {
    Filter1value1 = (double)StringToTime(_Filter1value1);
    Filter1value2 = (double)StringToTime(_Filter1value2);
  }
  else
  {
    Filter1value1 = StringToDouble(_Filter1value1);
    Filter1value2 = StringToDouble(_Filter1value2);
  }
  
  EventKillTimer();
  _defaultQuotesAdapter.reset();
  _defaultEngine.process(_selectorArray, _selectorField,
        _AggregatorType, _AggregatorField,
        _display,
        _SortBy,
        Filter1value1, Filter1value2);
  Comment("");
}


void OnTick()
{
}


/*
// APPENDIX I. example of candle formation analysis
class CustomQuotesRecord: public QuotesRecord
{
  public:
    CustomQuotesRecord(): QuotesRecord() {}
    CustomQuotesRecord(const MqlRates &rate): QuotesRecord(rate) {}
    
    virtual void fillCustomFields() override
    {
      const double open = get(FIELD_PRICE_OPEN);
      const double close = get(FIELD_PRICE_CLOSE);
      const double high = get(FIELD_PRICE_HIGH);
      const double low = get(FIELD_PRICE_LOW);

      set(FIELD_CUSTOM1, close > open ? (open - low) / _Point : (high - open) / _Point);
      set(FIELD_CUSTOM2, close > open ? (high - close) / _Point : (close - low) / _Point);
    }
    
    virtual string legend(const int index) const override
    {
      if(index == FIELD_CUSTOM1) return "Put Off shadow (post-Open)";
      else
      if(index == FIELD_CUSTOM2) return "Pull Back shadow (pre-Close)";
      
      return QuotesRecord::legend(index);
    }
};
*/
